from .nodes import *
from .values import QueryTerm

class Interpreter:
    def __init__(self, default_attrname: str):
        self.default_attrname = default_attrname
    
    def visit(self, node):
        # AttrNameNode => visit_AttrNameNode
        method_name = f"visit_{type(node).__name__}"
        method = getattr(self, method_name)
        result = method(node)
        if isinstance(result, list):
            result = flatten_list(result)
        return result
    
    def visit_list(self, node):
        out = []
        for n in node:
            if isinstance(n, list):
                n = self.visit_list(n)
            else:
                n = self.visit(n)
                if (not isinstance(n, list)) and (not isinstance(n, dict)):
                    n = n.value
            out.append(n)
        
        return out

    def visit_AssignAttrNode(self, node):
        qt = {
            'match': {},
            'not_match': {}
        }

        attr_name = f"{node.node_a.value}"
        attr_value = f"{node.node_b.value}"

        if node.operator == 'is':
            qt['match'][attr_name] = [ attr_value ]
        elif node.operator == 'is_not':
            qt['not_match'][attr_name] = [ attr_value ]
        
        return QueryTerm(qt)

    def visit_DefaultTokenNode(self, node):
        return QueryTerm({
            'match': {
                f"{self.default_attrname}": [ f"{node.value}" ]
            },
            'not_match': {}
        })

    def visit_EmptyTokenNode(self, node):
        return QueryTerm({})

    def visit_ConjoinAttrNode(self, node):
        qt = conjoin_dict(self.visit(node.node_a).value, self.visit(node.node_b).value)
        return QueryTerm(qt)
    
    def visit_QuantifyNode(self, node):
        quant_num = node.quantifier

        if isinstance(node.node_a, list):
            queryTerms = []
            for n in node.node_a:
                visited = self.visit(n)
                if isinstance(visited, list):
                    queryTerms.append(visited)
                else:
                    queryTerms.append(visited.value)
        else:
            visited = self.visit(node.node_a)
            if isinstance(visited, list):
                queryTerms = visited
            else:
                queryTerms = [ visited.value ]

        # Expand quantifier
        queryTerms =  queryTerms * quant_num        
        return queryTerms


    def visit_LabelNode(self, node):
        label = node.label
        qts = []

        if isinstance(node.node_a, list):
            for n in node.node_a:
                n = self.visit(n)
                if not isinstance(n, list):
                    n = n.value
                qts.append(n)
        else:
            qts = self.visit(node.node_a)
            if not isinstance(qts, list):
                qts = [ qts.value ]
        
        # Flatten data
        qts = flatten_list(qts)

        return add_label(qts, label)


################################
####### Helper functions #######
################################
def flatten_list(lst: list):
    out = []
    for elem in lst:
        if not isinstance(elem, list):
            out.append(elem)
        else:
            flatten = flatten_list(elem)
            out += flatten
    
    return out

def conjoin_dict(dict1, dict2):
    qt = {
        'match': {},
        'not_match': {}
    }

    # Add two dicts together
    for operator in qt:
        for d_ in [dict1, dict2]:
            for attr_name, attr_values in d_[operator].items():
                if attr_name not in qt[operator]:
                    qt[operator][attr_name] = set()
                for val in attr_values:
                    qt[operator][attr_name].add(val)
    
    # Remove duplicate values
    for operator in qt:
        for attr_name in qt[operator]:
            qt[operator][attr_name] = list(qt[operator][attr_name])
    
    return qt

def add_label(qts, label):
    if isinstance(qts, dict):
        return add_label_qt(qts, label)
    if isinstance(qts, list):
        return [ add_label_qt(qt, label) for qt in qts ]
    raise Exception(f"Unexpected data struct: {qts}")


def add_label_qt(qt, label):
    if '__label__' not in qt:
        qt['__label__'] = []
    if label not in qt['__label__']:
        qt['__label__'].append(label)
    return qt



from .lexer import Lexer
from .tokens import Token, TokenType
from itertools import product
from copy import deepcopy


def expand_quantifiers(tokens: list, max_quant: int=15):
    records = set()

    quantify_info = []
    for i, tk in enumerate(tokens):
        if tk.type.name == 'TOKEN_QUANTIFIER':
            min_ = tk.value[0]
            max_ = (tk.value[1] if tk.value[1] != 'inf' else max_quant) + 1
            quantify_info.append([i, 
                list(range(min_, max_))]
            )

    results = []
    for quant_set in product( *[l[1] for l in quantify_info] ):
        # Replace tokens for different quantifiers
        new_tokens = deepcopy(tokens)
        for i, num in enumerate(quant_set):
            tk_idx = quantify_info[i][0]
            new_tokens[tk_idx] = Token(TokenType.TOKEN_QUANTIFIER, num)

        # Remove quantifier of one
        new_tokens = [ tk for tk in new_tokens if not (tk.type.name == 'TOKEN_QUANTIFIER' and tk.value == 1) ] 

        record = str(new_tokens)
        if record not in records:
            results.append(new_tokens)
            records.add(record)

    return results



from typing import Union
from .lexer import Lexer
from .parser import Parser
from .interpreter import Interpreter
from .expand_quantifiers import expand_quantifiers


def parse(cql: str, default_attr:Union[str, int]="word", max_quant:int=15):
    """Parse a CQL string into lists of queries in JSON format


    Parameters
    ----------
    cql : str
        CQL query
    default_attr : Union[str, int], optional
        Default attribute for a token to generate if
        none is provided in the CQL, by default "word".
        E.g., "item" is equivalent to [word="item"] if
        `default_attr` is set to "word"
    max_quant : int, optional
        The maximum quantity of the quantifiers * and +, by default 15

    Returns
    -------
    list
        A list of queries, each of them consists of a list of query terms
        represented as a dictionary.
    """
    if cql.strip() == '':
        return []

    tokens = list(Lexer(cql).generate_tokens())
    parser = Parser(tokens)
    queries = expand_quantifiers(tokens, max_quant)

    values = []
    for query in queries:
        parser = Parser(query)
        tree = parser.parse()
        interpreter = Interpreter(default_attrname=default_attr)
        value = interpreter.visit(tree)
        if len(value) > 0:
            values.append(value)

    return values



from enum import Enum
from dataclasses import dataclass

class TokenType(Enum):
    ATTR_NAME           = 0
    ATTR_VALUE          = 1
    ATTR_RELATION       = 2
    ATTR_AND            = 3
    TOKEN_QUANTIFIER    = 4
    TOKEN_LABEL         = 5
    EMPTY_TOKEN         = 6
    DEFAULT_TOKEN       = 7
    LPAREN              = 8
    RPAREN              = 9

@dataclass
class Token:
    type: TokenType
    value: any = None
    
    def __repr__(self):
        return self.type.name + (f":{self.value}" if self.value is not None else "")





from dataclasses import dataclass

@dataclass
class AttrNameNode:
    value: str

    def __repr__(self):
        return f"{self.value}"

@dataclass
class AttrValueNode:
    value: str

    def __repr__(self):
        return f"{self.value}"

@dataclass
class AssignAttrNode:
    node_a: any
    operator: any
    node_b: any

    def __repr__(self):
        return f"({self.node_a} {self.operator} {self.node_b})"

@dataclass
class DefaultTokenNode:
    value: any

    def __repr__(self):
        return f"(DFTattr is {self.value})"

@dataclass
class EmptyTokenNode:
    def __repr__(self):
        return "EMPTY_TOKEN"

@dataclass
class ConjoinAttrNode:
    node_a: any
    node_b: any

    def __repr__(self):
        return f"({self.node_a} & {self.node_b})"

@dataclass
class QuantifyNode:
    node_a: any
    quantifier: int

    def __repr__(self):
        return f"{self.node_a}" + "{" + f"{self.quantifier}" "}"

@dataclass
class LabelNode:
    node_a: any
    label: any

    def __repr__(self):
        return f"{self.label}:{self.node_a}"



from .tokens import Token, TokenType

WHITESPACE = ' \t\n'
NAME_CHARS = set('0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_-.')
DIGITS = set('0123456789')
QUANTIFIERS = set('?+*')
ESCAPECHAR = chr(92)   # backslash

class Lexer:
    def __init__(self, text):
        self.text = iter(text.strip())
        self.current_char = None
        self.char_in_token_brackets = False
        self.char_in_attr_quotes = False
        self.char_in_quantifiers = False
        self.advance()
    
    def advance(self):
        try:
            self.current_char = next(self.text)
        except StopIteration:
            self.current_char = None
    
    def generate_tokens(self):
        while self.current_char is not None:
            # TOKEN open quote
            if (not self.char_in_token_brackets) and self.current_char == '[':
                self.char_in_token_brackets = True
                self.advance()
                if self.current_char == ']':
                    yield Token(TokenType.EMPTY_TOKEN, "EMPTY_TOKEN")
            # TOKEN closing quote
            elif self.char_in_token_brackets and self.current_char == ']':
                self.char_in_token_brackets = False
                self.advance()
            # ATTR_VALUE open quote
            elif (not self.char_in_attr_quotes) and self.current_char == '"':
                self.char_in_attr_quotes = True
                self.advance()
                if self.current_char == '"':
                    raise Exception("Empty ATTR_VALUE")
                else:
                    yield self.generate_attr_value()
                    self.advance()
                    self.char_in_attr_quotes = False
            # ATTR_RELATION
            elif self.char_in_token_brackets and (not self.char_in_attr_quotes) and (self.current_char == '!' or self.current_char == '='):
                yield self.generate_attr_relation()
            # ATTR_NAME
            elif self.char_in_token_brackets and (not self.char_in_attr_quotes) and self.current_char in NAME_CHARS:
                yield self.generate_attr_name()
            # ATTR_VALUE
            #elif self.char_in_attr_quotes:
            #    yield self.generate_attr_value()
            # Spaces between ATTR
            elif self.char_in_token_brackets and (not self.char_in_attr_quotes) and self.current_char in WHITESPACE:
                self.advance()
            # ATTR_AND
            elif self.char_in_token_brackets and (not self.char_in_attr_quotes) and self.current_char == '&':
                yield Token(TokenType.ATTR_AND)
                self.advance()
            # Whitespaces between tokens
            elif (not self.char_in_token_brackets) and (not self.char_in_attr_quotes) and self.current_char in WHITESPACE:
                self.advance()
            # TOKEN_LABEL
            elif (not self.char_in_token_brackets) and (not self.char_in_attr_quotes) and (not self.char_in_quantifiers) and self.current_char in NAME_CHARS:
                yield self.generate_token_label()
            # TOKEN_LABEL end
            elif (not self.char_in_token_brackets) and (not self.char_in_attr_quotes) and (not self.char_in_quantifiers) and self.current_char == ':':
                self.advance()
            # TOKEN_QUANTIFIER open quote
            elif (not self.char_in_token_brackets) and (not self.char_in_attr_quotes) and (not self.char_in_quantifiers) and self.current_char == '{':
                self.char_in_quantifiers = True
                self.advance()
            # TOKEN_QUANTIFIER closing quote
            elif (not self.char_in_token_brackets) and (not self.char_in_attr_quotes) and self.char_in_quantifiers and self.current_char == '}':
                self.char_in_quantifiers = False
                self.advance()
            # TOKEN_QUANTIFIER
            elif (not self.char_in_token_brackets) and (not self.char_in_attr_quotes) and self.char_in_quantifiers and self.current_char in DIGITS:
                yield self.generate_token_quantifier()
            elif (not self.char_in_token_brackets) and (not self.char_in_attr_quotes) and self.current_char in QUANTIFIERS:
                if self.current_char == '?':
                    yield Token(TokenType.TOKEN_QUANTIFIER, (0, 1))
                elif self.current_char == '+':
                    yield Token(TokenType.TOKEN_QUANTIFIER, (1, 'inf'))
                elif self.current_char == '*':
                    yield Token(TokenType.TOKEN_QUANTIFIER, (0, 'inf'))
                else:
                    raise Exception(f"Invalid character {self.current_char}")
                self.advance()
            # GROUP expressions
            elif (not self.char_in_token_brackets) and (not self.char_in_attr_quotes) and self.current_char == '(':
                yield Token(TokenType.LPAREN)
                self.advance()
            elif (not self.char_in_token_brackets) and (not self.char_in_attr_quotes) and self.current_char == ')':
                yield Token(TokenType.RPAREN)
                self.advance()
            else:
                raise Exception(f"Illegal character '{self.current_char}'")


    def generate_attr_name(self):
        name_str = self.current_char
        self.advance()

        while self.current_char != None and self.current_char in NAME_CHARS:
            name_str += self.current_char
            self.advance()

        return Token(TokenType.ATTR_NAME, name_str)


    def generate_attr_value(self):
        attr_str = ''

        #while self.current_char != None and self.current_char != '"':
        while self.current_char != None:
            if self.current_char == ESCAPECHAR:
                self.advance()
                if self.current_char is None:
                    raise Exception(f"Illegal character in generate_attr_value {self.current_char}")
                # Remove backslash escape for literal double quote
                elif self.current_char == '"':
                    attr_str += self.current_char
                # Show regex backslash escape
                else:
                    attr_str += ESCAPECHAR + self.current_char
                self.advance()
                continue
            
            if self.current_char == '"':
                break 
            
            attr_str += self.current_char
            self.advance()
        
        if self.char_in_token_brackets:
           return Token(TokenType.ATTR_VALUE, attr_str)
        else:
            return Token(TokenType.DEFAULT_TOKEN, attr_str)


    def generate_attr_relation(self):
        if self.current_char == '=':
            self.advance()
            return Token(TokenType.ATTR_RELATION, 'is')
        elif self.current_char == '!':
            self.advance()
            if self.current_char == '=':
                self.advance()
                return Token(TokenType.ATTR_RELATION, 'is_not')
        raise Exception(f"Illegal character '{self.current_char}'")


    def generate_token_quantifier(self):
        num_of_comma = 0
        q_min, q_max = None, None

        while self.current_char != None and (self.current_char in DIGITS or self.current_char == ',' or self.current_char in WHITESPACE):
            if self.current_char == ',':
                num_of_comma += 1
                if num_of_comma > 1: break
                self.advance()
            elif self.current_char in DIGITS:
                if q_min is None:
                    q_min = self.generate_int()
                    q_max = q_min
                else:
                    q_max = self.generate_int()
            elif self.current_char in WHITESPACE:
                self.advance()
        
        if q_min is None:
            raise Exception(f"Invalid character in quantifier: {self.current_char}")
        
        return Token(TokenType.TOKEN_QUANTIFIER, (q_min, q_max))


    def generate_token_label(self):
        label_str = self.current_char
        self.advance()

        while self.current_char != None and self.current_char in NAME_CHARS:
            label_str += self.current_char
            self.advance()
        
        return Token(TokenType.TOKEN_LABEL, label_str)


    def generate_int(self):
        num_str = self.current_char
        self.advance()

        while self.current_char != None and self.current_char in DIGITS:
            num_str += self.current_char
            self.advance()
        
        return int(num_str)


from .tokens import TokenType
from .nodes import *

class Parser:
    def __init__(self, tokens):
        self.tokens = iter(tokens)
        self.advance()

    def raise_error(self):
        raise Exception("Syntax Error")

    def advance(self):
        try:
            self.current_token = next(self.tokens)
        except StopIteration:
            self.current_token = None

    def parse(self):
        if self.current_token is None:
            return None

        result = self.list_wordGroups()

        if self.current_token != None:
            self.raise_error()

        return result

    def list_wordGroups(self):
        result = []
        while self.current_token != None:
            group = self.wordGroup()
            result.append(group)

        return result

    def wordGroup(self):
        token = self.current_token
        word_group = []
        label = None
        quant = None

        # Check whether group is labeled
        if token.type == TokenType.TOKEN_LABEL:
            label = token.value
            self.advance()

        # Parentheses as group
        if self.current_token.type == TokenType.LPAREN:
            self.advance()
            while self.current_token != None and self.current_token.type in {TokenType.DEFAULT_TOKEN, TokenType.EMPTY_TOKEN, TokenType.TOKEN_LABEL, TokenType.ATTR_NAME, TokenType.LPAREN, TokenType.TOKEN_LABEL}:
                if self.current_token.type == TokenType.LPAREN:
                    result = self.wordGroup()
                elif self.current_token.type == TokenType.TOKEN_LABEL:
                    result = self.wordGroup()
                else:
                    result = self.word()
                    #print(f"  {result}")
                    #print(f"  {self.current_token}")
                word_group.append(result)
            
            if self.current_token.type != TokenType.RPAREN:
                self.raise_error()
            self.advance()

            # Check trailing quantifier
            if self.current_token != None and self.current_token.type == TokenType.TOKEN_QUANTIFIER:
                quant = self.current_token.value
                self.advance()

            # Add label and quantifiers to group
            if quant != None:
                word_group = QuantifyNode(word_group, quant)
            if label != None:
                word_group = LabelNode(word_group, label)

            return word_group

        # Single word
        elif self.current_token.type in {TokenType.DEFAULT_TOKEN, TokenType.EMPTY_TOKEN, TokenType.ATTR_NAME}:
            #print(self.current_token)
            word_group = self.word()
            #print(word_group)
            if label != None:
                word_group = LabelNode(word_group, label)

            return word_group


    def word(self):
        quant = None
        # Directly return DEFAULT/EMPTY_TOKEN
        if self.current_token.type in {TokenType.DEFAULT_TOKEN, TokenType.EMPTY_TOKEN}:
            if self.current_token.type == TokenType.DEFAULT_TOKEN:
                word = DefaultTokenNode(self.current_token.value)
            else:
                word = EmptyTokenNode()

            # Check trailing quantifier
            self.advance()
            if self.current_token != None and self.current_token.type == TokenType.TOKEN_QUANTIFIER:
                quant = self.current_token.value
                self.advance()

            if quant != None:
                word = QuantifyNode(word, quant)

            return word

        # Construct word from attributes
        word = self.word_attrpair()

        while self.current_token != None and self.current_token.type == TokenType.ATTR_AND:
            self.advance()
            word = ConjoinAttrNode(word, self.word_attrpair())

        # Check quantifier
        if self.current_token != None and self.current_token.type == TokenType.TOKEN_QUANTIFIER:
            quant = self.current_token.value
            self.advance()
        if quant != None:
            word = QuantifyNode(word, quant)

        return word


    def word_attrpair(self):
        token = self.current_token

        if token.type == TokenType.ATTR_NAME:
            self.advance()
            if self.current_token.type == TokenType.ATTR_RELATION:
                operator = self.current_token.value
                self.advance()
                if self.current_token.type == TokenType.ATTR_VALUE:
                    result = AssignAttrNode(
                        AttrNameNode(token.value),
                        operator,
                        AttrValueNode(self.current_token.value))
                    self.advance()
                    return result

        self.raise_error()



from setuptools import setup

with open("README.md", encoding="utf-8") as f:
      long_description = f.read().strip()

setup(name='cqls',
      version='0.1.5',
      description='A parser for an often-used subset of Corpus Query Language',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://github.com/liao961120/cqls',
      author='Yongfu Liao',
      author_email='liao961120@github.com',
      license='MIT',
      packages=['cqls'],
      tests_require=['deepdiff'],
      zip_safe=False)
